#include "../../SourceExt/angelscript/include/angelscript.h"
#include <cstdio>
#include <cassert>

// --------------------------------------------------------------------------------------------------------------------
struct TColor
{
  float r, g, b;

public:
  constexpr TColor() noexcept : r{0.0f}, g{0.0f}, b{0.0f} {}
  constexpr TColor( const TColor&) noexcept = default;
  constexpr TColor( float ir, float ig, float ib) noexcept : r{ir}, g{ig}, b{ib} {}

  TColor& operator = (const TColor&) noexcept = default;
};

inline constexpr TColor operator * (const TColor& c, float scale) noexcept { return TColor{ c.r*scale, c.g*scale, c.b*scale }; }

struct ScriptHelper {
  static void ConstructTColor( TColor *poF) { new(poF) TColor; }
  static void DestroyTColor( TColor *poF) { poF->~TColor(); }
};

const TColor& GetColor() { static TColor c{ 4.0f, 5.0f, 6.0f }; return c; }
void SetColor( const TColor& c) { printf( "%.2f\n", c.r); }
void Log( const asSMessageInfo* info) { printf( "L%d: %s\n", info->row, info->message); }

// --------------------------------------------------------------------------------------------------------------------
int main( int argc, char** argv)
{
  // ******************************************************************************************************************
  // ******** This is it. Remove the manual TColor( ...) copy and it works. Leave it and TColor is passed by value 
  // ******** to the operator taking a reference
  const char* scriptCode = 
    "void scriptFunction() {\n"
    "  SetColor( TColor( GetColor()) * 0.1f);\n"
    "}\n";

  asIScriptEngine* engine = asCreateScriptEngine( ANGELSCRIPT_VERSION);
  assert( engine );
  engine->SetMessageCallback( asFUNCTION( Log), nullptr, asCALL_CDECL);
  int r = engine->SetEngineProperty( asEP_ALLOW_UNSAFE_REFERENCES, 1); assert( r >= 0 );

  engine->RegisterObjectType( "TColor", sizeof(TColor), asOBJ_VALUE | asOBJ_APP_CLASS_C);
  engine->RegisterObjectBehaviour( "TColor", asBEHAVE_CONSTRUCT, "void f()", asFUNCTIONPR( ScriptHelper::ConstructTColor, (TColor *), void), asCALL_CDECL_OBJLAST);
  engine->RegisterObjectBehaviour( "TColor", asBEHAVE_DESTRUCT, "void f()", asFUNCTIONPR( ScriptHelper::DestroyTColor, (TColor *), void), asCALL_CDECL_OBJLAST);
  engine->RegisterObjectProperty( "TColor", "float r", offsetof( TColor, r));
  engine->RegisterObjectProperty( "TColor", "float g", offsetof( TColor, g));
  engine->RegisterObjectProperty( "TColor", "float b", offsetof( TColor, b));
  engine->RegisterObjectMethod( "TColor", "TColor& opAssign(const TColor& in)", asMETHODPR( TColor, operator =, (const TColor&), TColor&), asCALL_THISCALL);
  engine->RegisterObjectMethod( "TColor", "TColor opMul( float) const", asFUNCTIONPR( operator *, (const TColor&, float), TColor), asCALL_CDECL_OBJFIRST);

  engine->RegisterGlobalFunction( "const TColor& GetColor()", asFUNCTION( GetColor), asCALL_CDECL);
  engine->RegisterGlobalFunction( "void SetColor( const TColor& in)", asFUNCTION( SetColor), asCALL_CDECL);

  auto mymod = engine->GetModule( "mymod", asGM_ALWAYS_CREATE);
  assert( mymod );
  r = mymod->AddScriptSection( "name", scriptCode); assert( r >= 0 );
  r = mymod->Build(); assert( r >= 0 );

  auto scriptFunc = mymod->GetFunctionByName( "scriptFunction");
  assert( scriptFunc );

  auto myctx = engine->CreateContext();
  r = myctx->Prepare( scriptFunc); assert( r >= 0 );
  r = myctx->Execute(); assert( r >= 0 );
}
